/**
 * Copyright (c) 2012 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#include "sdk_common.h"
#include "ble.h"
#include "ble_tj.h"
#include "ble_srv_common.h"


#define BLE_UUID_TJ_COMMAND_CHARACTERISTIC	0x0002                     	   /**< The UUID of the RX Command Characteristic. */
#define BLE_UUID_TJ_AUTH_CHARACTERISTIC		0x0003                     	   /**< The UUID of the RX Auth Characteristic. */
#define BLE_UUID_TJ_TX_RSP_CHARACTERISTIC   0x0004                     	   /**< The UUID of the TX Response Characteristic. */
#define BLE_UUID_TJ_TX_EVENT_CHARACTERISTIC 0x0005                     	   /**< The UUID of the TX Response Characteristic. */

#define BLE_TJ_MAX_RX_CHAR_LEN         (NRF_SDH_BLE_GATT_MAX_MTU_SIZE - OPCODE_LENGTH - HANDLE_LENGTH)         /**< Maximum length of the RX Characteristic (in bytes). */
#define BLE_NUS_MAX_TX_CHAR_LEN         (NRF_SDH_BLE_GATT_MAX_MTU_SIZE - OPCODE_LENGTH - HANDLE_LENGTH)         /**< Maximum length of the TX Characteristic (in bytes). */

#define TJ_BASE_UUID                  {{0x9E, 0xCA, 0xDC, 0x24, 0x0E, 0xE5, 0xA9, 0xE0, 0x93, 0xF3, 0xA3, 0xB5, 0x00, 0x00, 0x56, 0x78}} /**< Used vendor specific UUID. */


/**@brief Function for handling the @ref BLE_GAP_EVT_CONNECTED event from the SoftDevice.
 *
 * @param[in] p_nus     TJ Service structure.
 * @param[in] p_ble_evt Pointer to the event received from BLE stack.
 */
static void on_tj_connect(ble_tj_t * p_tj, ble_evt_t const * p_ble_evt)
{
    p_tj->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}


/**@brief Function for handling the @ref BLE_GAP_EVT_DISCONNECTED event from the SoftDevice.
 *
 * @param[in] p_nus     Nordic UART Service structure.
 * @param[in] p_ble_evt Pointer to the event received from BLE stack.
 */
static void on_tj_disconnect(ble_tj_t * p_tj, ble_evt_t const * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_tj->conn_handle = BLE_CONN_HANDLE_INVALID;
}


/**@brief Function for handling the @ref BLE_GATTS_EVT_WRITE event from the SoftDevice.
 *
 * @param[in] p_nus     Nordic UART Service structure.
 * @param[in] p_ble_evt Pointer to the event received from BLE stack.
 */
static void on_tj_write(ble_tj_t * p_tj, ble_evt_t const * p_ble_evt)
{
    ble_gatts_evt_write_t const * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;
    ble_tj_evt_t evt;
    evt.p_tj = p_tj;

	// Register for notifications on event/response to commands
    if (   (p_evt_write->handle == p_tj->tx_rsp_handles.cccd_handle)
        && (p_evt_write->len == 2))
    {
    	// Client registers to notifications on rsp
        if (ble_srv_is_notification_enabled(p_evt_write->data))
        {
            p_tj->is_rsp_notification_enabled = true;
            evt.type = BLE_TJ_EVT_RSP_STARTED;
        }
        else
        {
            p_tj->is_rsp_notification_enabled = false;
            evt.type = BLE_TJ_EVT_RSP_STOPPED;
        }
        p_tj->data_handler(&evt);
    }
	else if (   (p_evt_write->handle == p_tj->tx_event_handles.cccd_handle)
        && (p_evt_write->len == 2))
    {
    	// Client registers to notifications on event
        if (ble_srv_is_notification_enabled(p_evt_write->data))
        {
            p_tj->is_event_notification_enabled = true;
            evt.type = BLE_TJ_EVT_EVENT_STARTED;
        }
        else
        {
            p_tj->is_event_notification_enabled = false;
            evt.type = BLE_TJ_EVT_EVENT_STOPPED;
        }
        p_tj->data_handler(&evt);
    }
    else if (   (p_evt_write->handle == p_tj->rx_command_handles.value_handle)
             && (p_tj->data_handler != NULL))
    {
        evt.params.rx_data.p_data = p_evt_write->data;
        evt.params.rx_data.length = p_evt_write->len;
        evt.type = BLE_TJ_EVT_RX_DATA;
        p_tj->data_handler(&evt);
    }
	else if (   (p_evt_write->handle == p_tj->rx_auth_handles.value_handle)
             && (p_tj->data_handler != NULL))
    {
        evt.params.rx_data.p_data = p_evt_write->data;
        evt.params.rx_data.length = p_evt_write->len;
        evt.type = BLE_TJ_EVT_RX_AUTH;
        p_tj->data_handler(&evt);
    }
    else
    {
        // Do Nothing. This event is not relevant for this service.
    }
}


/**@brief Function for adding TX characteristic.
 *
 * @param[in] p_nus       Nordic UART Service structure.
 * @param[in] p_nus_init  Information needed to initialize the service.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t tx_rsp_add(ble_tj_t * p_tj, ble_tj_init_t const * p_tj_init)
{
    /**@snippet [Adding proprietary characteristic to the SoftDevice] */
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_md_t cccd_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&cccd_md, 0, sizeof(cccd_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

    cccd_md.vloc = BLE_GATTS_VLOC_STACK;

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.notify = 1;
	char_md.char_props.read	  = 1;
    char_md.p_char_user_desc  = NULL;
    char_md.p_char_pf         = NULL;
    char_md.p_user_desc_md    = NULL;
    char_md.p_cccd_md         = &cccd_md;
    char_md.p_sccd_md         = NULL;

    ble_uuid.type = p_tj->uuid_type;
    ble_uuid.uuid = BLE_UUID_TJ_TX_RSP_CHARACTERISTIC;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

    attr_md.vloc    = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 0;
    attr_md.wr_auth = 0;
    attr_md.vlen    = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = sizeof(uint8_t);
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = BLE_NUS_MAX_TX_CHAR_LEN;

    return sd_ble_gatts_characteristic_add(p_tj->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_tj->tx_rsp_handles);
    /**@snippet [Adding proprietary characteristic to the SoftDevice] */
}


/**@brief Function for adding TX characteristic.
 *
 * @param[in] p_nus       Nordic UART Service structure.
 * @param[in] p_nus_init  Information needed to initialize the service.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t tx_event_add(ble_tj_t * p_tj, ble_tj_init_t const * p_tj_init)
{
    /**@snippet [Adding proprietary characteristic to the SoftDevice] */
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_md_t cccd_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&cccd_md, 0, sizeof(cccd_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

    cccd_md.vloc = BLE_GATTS_VLOC_STACK;

    memset(&char_md, 0, sizeof(char_md));

    char_md.char_props.notify = 1;
	char_md.char_props.read	  = 1;
    char_md.p_char_user_desc  = NULL;
    char_md.p_char_pf         = NULL;
    char_md.p_user_desc_md    = NULL;
    char_md.p_cccd_md         = &cccd_md;
    char_md.p_sccd_md         = NULL;

    ble_uuid.type = p_tj->uuid_type;
    ble_uuid.uuid = BLE_UUID_TJ_TX_EVENT_CHARACTERISTIC;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

    attr_md.vloc    = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 0;
    attr_md.wr_auth = 0;
    attr_md.vlen    = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = sizeof(uint8_t);
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = BLE_NUS_MAX_TX_CHAR_LEN;

    return sd_ble_gatts_characteristic_add(p_tj->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_tj->tx_event_handles);
    /**@snippet [Adding proprietary characteristic to the SoftDevice] */
}


/**@brief Function for adding RX characteristic.
 *
 * @param[in] p_nus       TJ Service structure.
 * @param[in] p_nus_init  Information needed to initialize the service.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t rx_command_char_add(ble_tj_t * p_tj, const ble_tj_init_t * p_tj_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&char_md, 0, sizeof(char_md));

	char_md.char_props.read          = 1;
    char_md.char_props.write         = 1;
    char_md.char_props.write_wo_resp = 1;
    char_md.p_char_user_desc         = NULL;
    char_md.p_char_pf                = NULL;
    char_md.p_user_desc_md           = NULL;
    char_md.p_cccd_md                = NULL;
    char_md.p_sccd_md                = NULL;

    ble_uuid.type = p_tj->uuid_type;
    ble_uuid.uuid = BLE_UUID_TJ_COMMAND_CHARACTERISTIC;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

    attr_md.vloc    = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 0;
    attr_md.wr_auth = 0;
    attr_md.vlen    = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = 1;
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = BLE_TJ_MAX_RX_CHAR_LEN;

    return sd_ble_gatts_characteristic_add(p_tj->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_tj->rx_command_handles);
}



/**@brief Function for adding RX characteristic.
 *
 * @param[in] p_nus       TJ Service structure.
 * @param[in] p_nus_init  Information needed to initialize the service.
 *
 * @return NRF_SUCCESS on success, otherwise an error code.
 */
static uint32_t rx_auth_char_add(ble_tj_t * p_tj, const ble_tj_init_t * p_tj_init)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&char_md, 0, sizeof(char_md));

	char_md.char_props.read          = 1;
    char_md.char_props.write         = 1;
    char_md.char_props.write_wo_resp = 1;
    char_md.p_char_user_desc         = NULL;
    char_md.p_char_pf                = NULL;
    char_md.p_user_desc_md           = NULL;
    char_md.p_cccd_md                = NULL;
    char_md.p_sccd_md                = NULL;

    ble_uuid.type = p_tj->uuid_type;
    ble_uuid.uuid = BLE_UUID_TJ_AUTH_CHARACTERISTIC;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);

    attr_md.vloc    = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth = 0;
    attr_md.wr_auth = 0;
    attr_md.vlen    = 1;

    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid    = &ble_uuid;
    attr_char_value.p_attr_md = &attr_md;
    attr_char_value.init_len  = 1;
    attr_char_value.init_offs = 0;
    attr_char_value.max_len   = BLE_TJ_MAX_RX_CHAR_LEN;

    return sd_ble_gatts_characteristic_add(p_tj->service_handle,
                                           &char_md,
                                           &attr_char_value,
                                           &p_tj->rx_auth_handles);
}



void ble_tj_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context)
{
    if ((p_context == NULL) || (p_ble_evt == NULL))
    {
        return;
    }

    ble_tj_t * p_tj = (ble_tj_t *)p_context;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_tj_connect(p_tj, p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            on_tj_disconnect(p_tj, p_ble_evt);
            break;

        case BLE_GATTS_EVT_WRITE:
            on_tj_write(p_tj, p_ble_evt);
            break;


        case BLE_GATTS_EVT_HVN_TX_COMPLETE:
        {
            //notify with empty data that some tx was completed.
            ble_tj_evt_t evt = {
                    .type = BLE_TJ_EVT_TX_RDY,
                    .p_tj = p_tj
            };
            p_tj->data_handler(&evt);
            break;
        }
        default:
            // No implementation needed.
            break;
    }
}


uint32_t ble_tj_init(ble_tj_t * p_tj, ble_tj_init_t const * p_tj_init)
{
    uint32_t      err_code;
    ble_uuid_t    ble_uuid;
    ble_uuid128_t tj_base_uuid = TJ_BASE_UUID;

    VERIFY_PARAM_NOT_NULL(p_tj);
    VERIFY_PARAM_NOT_NULL(p_tj_init);

    // Initialize the service structure.
    p_tj->conn_handle             = BLE_CONN_HANDLE_INVALID;
    p_tj->data_handler            = p_tj_init->data_handler;
    p_tj->is_rsp_notification_enabled = false;
    p_tj->is_event_notification_enabled = false;

    /**@snippet [Adding proprietary Service to the SoftDevice] */
    // Add a custom base UUID.
    err_code = sd_ble_uuid_vs_add(&tj_base_uuid, &p_tj->uuid_type);
    VERIFY_SUCCESS(err_code);

    ble_uuid.type = p_tj->uuid_type;
    ble_uuid.uuid = BLE_UUID_TJ_SERVICE;

    // Add the service.
    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                        &ble_uuid,
                                        &p_tj->service_handle);
    /**@snippet [Adding proprietary Service to the SoftDevice] */
    VERIFY_SUCCESS(err_code);

    // Add the RX Command Characteristic.
    err_code = rx_command_char_add(p_tj, p_tj_init);
    VERIFY_SUCCESS(err_code);

	// Add the rx authentication characteristic
  	err_code = rx_auth_char_add(p_tj, p_tj_init);
   	VERIFY_SUCCESS(err_code);

    // Add the TX RSP Characteristic.
    err_code = tx_rsp_add(p_tj, p_tj_init);
    VERIFY_SUCCESS(err_code);

	
    // Add the TX Alert Characteristic.
  	err_code = tx_event_add(p_tj, p_tj_init);
    VERIFY_SUCCESS(err_code);

    return NRF_SUCCESS;
}


uint32_t ble_tj_rsp_send(ble_tj_t * p_tj, uint8_t * p_data, uint16_t * p_length)
{
    ble_gatts_hvx_params_t hvx_params;

    VERIFY_PARAM_NOT_NULL(p_tj);

    if ((p_tj->conn_handle == BLE_CONN_HANDLE_INVALID) || (!p_tj->is_rsp_notification_enabled))
    {
        return NRF_ERROR_INVALID_STATE;
    }

    if (*p_length > BLE_TJ_MAX_DATA_LEN)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

    memset(&hvx_params, 0, sizeof(hvx_params));

    hvx_params.handle = p_tj->tx_rsp_handles.value_handle;
    hvx_params.p_data = p_data;
    hvx_params.p_len  = p_length;
    hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;

    return sd_ble_gatts_hvx(p_tj->conn_handle, &hvx_params);
}

uint32_t ble_tj_event_send(ble_tj_t * p_tj, uint8_t * p_event, uint16_t * p_length)
{
    ble_gatts_hvx_params_t hvx_params;

    VERIFY_PARAM_NOT_NULL(p_tj);

    if (p_tj->conn_handle == BLE_CONN_HANDLE_INVALID)
    {
        return NRF_ERROR_INVALID_STATE;
    }

    if (*p_length > BLE_TJ_MAX_DATA_LEN)
    {
        return NRF_ERROR_INVALID_PARAM;
    }

	if (!p_tj->is_event_notification_enabled)
	{
		return 0;
	}

    memset(&hvx_params, 0, sizeof(hvx_params));

    hvx_params.handle = p_tj->tx_event_handles.value_handle;
    hvx_params.p_data = p_event;
    hvx_params.p_len  = p_length;
    hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;

    return sd_ble_gatts_hvx(p_tj->conn_handle, &hvx_params);
}

